import datetime
import os
import re
from typing import Optional, Union, List

import pydantic
import ujson
from deta import Deta
from pydantic import Field, BaseModel

from odetam.exceptions import NoProjectKey, DetaError, ItemNotFound
from odetam.field import DetaField
from odetam.query import DetaQuery, DetaQueryStatement, DetaQueryList

# TODO: Add serialization and comparison code that turns dates into numbers so that
# meaninful comparison can be performed

DETA_BASIC_TYPES = [dict, list, str, int, float, bool]
DETA_OPTIONAL_TYPES = [Optional[type_] for type_ in DETA_BASIC_TYPES]
DETA_BASIC_LIST_TYPES = [
    List[type_] for type_ in DETA_BASIC_TYPES + DETA_OPTIONAL_TYPES
]
DETA_TYPES = DETA_BASIC_TYPES + DETA_OPTIONAL_TYPES + DETA_BASIC_LIST_TYPES


class DetaModelMetaClass(pydantic.main.ModelMetaclass):
    def __new__(mcs, name, bases, dct):
        cls = super().__new__(mcs, name, bases, dct)
        cls.__db_name__ = re.sub(r"(?<!^)(?=[A-Z])", "_", name).lower()
        cls._db = None

        for name, field in cls.__fields__.items():
            setattr(cls, name, DetaField(field=field))
        return cls

    @property
    def __db__(cls):
        if cls._db:
            return cls._db
        try:
            deta = Deta(os.getenv("PROJECT_KEY"))
        except (AttributeError, AssertionError):
            raise NoProjectKey(
                "Ensure that the 'PROJECT_KEY' environment variable is set to your "
                "project key"
            )
        cls._db = deta.Base(cls.__db_name__)
        return cls._db


class DetaModel(BaseModel, metaclass=DetaModelMetaClass):
    __db__ = None
    key: Optional[str] = Field(
        None, title="Key", description="Primary key in the database"
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.key = self.key or None

    @classmethod
    def get(cls, key):
        """
        Get a single instance
        :param key: Deta database key
        :return: object found in database serialized into its pydantic object

        :raises ItemNotFound: No matching item was found
        """
        item = cls.__db__.get(key)
        if item is None:
            raise ItemNotFound("Could not find item matching that key")
        return cls._deserialize(item)

    @classmethod
    def get_all(cls):
        """Get all the records from the database"""
        records = cls.__db__.fetch().items
        return [cls._deserialize(record) for record in records]

    @classmethod
    def query(
        cls, query_statement: Union[DetaQuery, DetaQueryStatement, DetaQueryList]
    ):
        """Get items from database based on the query."""
        found = cls.__db__.fetch(query_statement.as_query()).items
        return [cls._deserialize(item) for item in found]

    @classmethod
    def delete_key(cls, key):
        """Delete an item based on the key"""
        cls.__db__.delete(key)

    @classmethod
    def put_many(cls, items):
        """Put multiple instances at once

        :param items: List of pydantic objects to put in the database
        :returns: List of items successfully added, serialized with pydantic
        """
        records = []
        processed = []
        for item in items:
            exclude = set()
            if item.key is None:
                exclude = {"key"}
            records.append(item._serialize())
            if len(records) == 25:
                result = cls.__db__.put_many(records)
                processed.extend(result["processed"]["items"])
                records = []
        if records:
            result = cls.__db__.put_many(records)
            processed.extend(result["processed"]["items"])
        return [cls._deserialize(rec) for rec in processed]

    @classmethod
    def _db_put(cls, data):
        return cls.__db__.put(data)

    def save(self):
        """Saves the record to the database. Behaves as upsert, will create
        if not present. Database key will then be set on the object."""
        # exclude = set()
        # if self.key is None:
        #     exclude.add("key")
        # # this is dumb, but it ensures everything is in a json-serializable form
        # data = ujson.loads(self.json(exclude=exclude))
        saved = self._db_put(self._serialize())
        self.key = saved["key"]

    def delete(self):
        """Delete the open object from the database. The object will still exist in
        python, but will be deleted from the database and the key attribute will be
        set to None."""
        if not self.key:
            raise DetaError("Item does not have key for deletion")
        self.delete_key(self.key)
        self.key = None

    def _serialize(self):
        as_dict = {}
        for field_name, field in self.__class__.__fields__.items():
            if field_name == "key" and not self.key:
                continue
            if field.type_ in DETA_TYPES:
                as_dict[field_name] = getattr(self, field_name)
            elif field.type_ == datetime.datetime:
                as_dict[field_name] = getattr(self, field_name).timestamp()
            elif field.type_ == datetime.date:
                as_dict[field_name] = int(getattr(self, field_name).strftime("%Y%m%d"))
            elif field.type_ == datetime.time:
                as_dict[field_name] = int(
                    getattr(self, field_name).strftime("%H%M%S%f")
                )
            else:
                as_dict[field_name] = ujson.loads(self.json(include={field_name}))[
                    field_name
                ]
        return as_dict

    @classmethod
    def _deserialize(cls, data):
        as_dict = {}
        for field_name, field in cls.__fields__.items():
            if not data.get(field_name):
                continue
            if field.type_ in DETA_TYPES:
                as_dict[field_name] = data[field_name]
            elif field.type_ == datetime.datetime:
                as_dict[field_name] = datetime.datetime.fromtimestamp(data[field_name])
            elif field.type_ == datetime.date:
                as_dict[field_name] = datetime.datetime.strptime(
                    str(data[field_name]), "%Y%m%d"
                ).date()
            elif field.type_ == datetime.time:
                as_dict[field_name] = datetime.datetime.strptime(
                    str(data[field_name]), "%H%M%S%f"
                ).time()
            else:
                as_dict[field_name] = ujson.loads(data[field_name])
        return cls.parse_obj(as_dict)



import datetime
from typing import List

import ujson
from pydantic import BaseModel

from odetam.exceptions import InvalidDetaQuery
from odetam.query import DetaQuery

NON_STR_TYPES = [
    dict,
    int,
    float,
    bool,
    List[dict],
    List[float],
    List[int],
    List[float],
    List[bool],
    datetime.datetime,
    datetime.date,
    datetime.time,
]


def _handle_datetimes(possible_datetime):
    if isinstance(possible_datetime, datetime.datetime):
        return possible_datetime.timestamp()
    if isinstance(possible_datetime, datetime.date):
        return int(possible_datetime.strftime("%Y%m%d"))
    if isinstance(possible_datetime, datetime.time):
        return int(possible_datetime.strftime("%H%M%S%f"))
    return possible_datetime


class DetaField:
    def __init__(self, field):
        self.field = field

    def _check_type(self, other):
        if not isinstance(other, self.field.type_):
            raise TypeError("Cannot compare different types")

    def _query_expression(self, operator, data):
        if isinstance(data, BaseModel):
            data = ujson.loads(data.json())
        return DetaQuery(condition=f"{self.field.name}?{operator}", value=data)

    def __eq__(self, other):
        return DetaQuery(condition=self.field.name, value=_handle_datetimes(other))

    def __ne__(self, other):
        self._check_type(other)
        return self._query_expression("ne", _handle_datetimes(other))

    def __lt__(self, other):
        self._check_type(other)
        return self._query_expression("lt", _handle_datetimes(other))

    def __gt__(self, other):
        self._check_type(other)
        return self._query_expression("gt", _handle_datetimes(other))

    def __le__(self, other):
        self._check_type(other)
        return self._query_expression("lte", _handle_datetimes(other))

    def __ge__(self, other):
        self._check_type(other)
        return self._query_expression("gte", _handle_datetimes(other))

    def prefix(self, other):
        if not isinstance(other, str) or self.field.type_ != str:
            raise InvalidDetaQuery("Prefix is only valid for string types")
        return self._query_expression("pfx", other)

    def range(self, lower, upper):
        if self.field.type_ in (datetime.date, datetime.time, datetime.datetime):
            lower = _handle_datetimes(lower)
            upper = _handle_datetimes(upper)
        if self.field.type_ not in (
            int,
            float,
            datetime.date,
            datetime.time,
            datetime.datetime,
        ):
            raise InvalidDetaQuery("Range is only valid for number types")
        if not isinstance(lower, (int, float)):
            raise InvalidDetaQuery("Lower must be a number")
        if not isinstance(upper, (int, float)):
            raise InvalidDetaQuery("Upper must be a number")
        if upper <= lower:
            raise InvalidDetaQuery("Lower must be less than upper")
        return DetaQuery(condition=f"{self.field.name}?r", value=[lower, upper])

    def contains(self, other: str):
        if not isinstance(other, str) or self.field.type_ in NON_STR_TYPES:
            raise InvalidDetaQuery(
                "Contains is only valid for strings or lists of strings"
            )
        return DetaQuery(condition=f"{self.field.name}?contains", value=other)

    def not_contains(self, other: str):
        if not isinstance(other, str) or self.field.type_ in NON_STR_TYPES:
            raise InvalidDetaQuery(
                "Not contains is only valid for strings or lists of strings"
            )
        return DetaQuery(condition=f"{self.field.name}?not_contains", value=other)



from typing import List

from odetam.exceptions import InvalidDetaQuery


# TODO: Add better type checking to avoid errors


class DetaQuery:
    def __init__(self, condition, value):
        self.condition = condition
        self.value = value

    def __and__(self, other):
        if isinstance(other, DetaQueryStatement):
            return other & self
        return DetaQueryStatement([self, other])

    def __or__(self, other):
        if isinstance(other, DetaQueryList):
            return other | self
        return DetaQueryList(conditions=[self, other])

    def as_query(self):
        return {self.condition: self.value}


class DetaQueryList:
    def __init__(self, conditions):
        self.conditions = conditions

    def __or__(self, other):
        if isinstance(other, DetaQueryList):
            self.conditions.extend(other.conditions)
        else:
            self.conditions.append(other)
        return self

    def as_query(self):
        return [query.as_query() for query in self.conditions]


class DetaQueryStatement:
    def __init__(self, conditions: List[DetaQuery]):
        self.conditions = conditions

    def __and__(self, other):
        if isinstance(other, DetaQuery):
            self.conditions.append(other)
        if isinstance(other, DetaQueryStatement):
            self.conditions.extend(other.conditions)
        if isinstance(other, DetaQueryList):
            raise InvalidDetaQuery("Incorrect and/or nesting")
        return self

    def __or__(self, other):
        if isinstance(other, DetaQueryList):
            other.conditions.append(self)
            return other
        return DetaQueryList(conditions=[self, other])

    def as_query(self):
        return {query.condition: query.value for query in self.conditions}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['odetam']

package_data = \
{'': ['*']}

install_requires = \
['deta>=1.0,<2.0', 'pydantic>=1.7,<2.0', 'ujson>=4.0,<5.0']

setup_kwargs = {
    'name': 'odetam',
    'version': '1.1.0a1',
    'description': 'A simple ODM (Object Document Mapper) for Deta Base, based on pydantic.',
    'long_description': '# ODetaM\n\n[![Build Status](https://travis-ci.org/rickh94/ODetaM.svg?branch=main)](https://travis-ci.org/rickh94/ODetaM)\n[![codecov](https://codecov.io/gh/rickh94/odetam/branch/main/graph/badge.svg?token=BLDIMHU9FB)](https://codecov.io/gh/rickh94/odetam)\n\nA simple ODM (Object Document Mapper) for [Deta Base](https://deta.sh) base on\n[pydantic](https://github.com/samuelcolvin/pydantic/).\n\n## Installation\n\n`pip install odetam`\n\n## Usage\n\nCreate pydantic models as normal, but inherit from `DetaModel` instead of pydantic\nBaseModel. You will need to set the environment variable `PROJECT_KEY` to your Deta\nproject key so that databases can be accessed/created. This is a secret key, so handle\nit appropriately (hence the environment variable). Intended for use with FastAPI, but\nthe Deta API is not asynchronous, so any framework could potentially be used.\n\nBases will be automatically created based on model names (changed from\nPascalCase/CamelCase case to snake_case). A `key` field (Deta\'s unique id) will be\nautomatically added to any model. You can supply the key on creation, or Deta will\ngenerate one automatically and it will be added to the object when it is saved.\n\n### Get All\n\nDetaModel.get_all() may be broken for large bases. The main deta python library has\nimplemented some breaking changes (though definite improvements) around the fetching\nof records from the Base. They have implemented a new automatic pagination system \nthat I have not had time to play with yet. Queries are recommended!\n\n## Example\n\n```python\nimport datetime\nfrom typing import List\n\nfrom odetam import DetaModel\n\n\nclass Captain(DetaModel):\n    name: str\n    joined: datetime.date\n    ships: List[str]\n\n\n# create\nkirk = Captain(\n    name="James T. Kirk",\n    joined=datetime.date(2252, 1, 1),\n    ships=["Enterprise"],\n)\n\nsisko = Captain(\n    name="Benjamin Sisko",\n    joined=datetime.date(2350, 1, 1),\n    ships=["Deep Space 9", "Defiant"],\n)\n\n# initial save, key is now set\nkirk.save()\n\n# update the object\nkirk.ships.append("Enterprise-A")\n\n# save again, this will be an update\nkirk.save()\n\nsisko.save()\n\nCaptain.get_all()\n# [\n#     Captain(\n#         name="James T. Kirk", \n#         joined=datetime.date(2252, 01, 01), \n#         ships=["Enterprise", "Enterprise-A"],\n#         key="key1",\n#     ),\n#     Captain(\n#         name="Benjamin Sisko",\n#         joined=datetime.date(2350, 01, 01), \n#         ships=["Deep Space 9", "Defiant"],\n#         key="key2",\n#     ),\n# ]\n\nCaptain.get("key1")\n# Captain(\n#     name="James T. Kirk", \n#     joined=datetime.date(2252, 01, 01), \n#     ships=["Enterprise", "Enterprise-A"],\n#     key="key1",\n# )\n\nCaptain.query(Captain.name == "James T. Kirk")\n# Captain(\n#     name="James T. Kirk", \n#     joined=datetime.date(2252, 01, 01), \n#     ships=["Enterprise", "Enterprise-A"],\n#     key="key1",\n# )\n\nCaptain.query(Captain.ships.contains("Defiant"))\n# Captain(\n#     name="Benjamin Sisko",\n#     joined=datetime.date(2350, 01, 01),\n#     ships=["Deep Space 9", "Defiant"],\n# )\n\nCaptain.query(Captain.name.prefix("Ben"))\n# Captain(\n#     name="Benjamin Sisko",\n#     joined=datetime.date(2350, 01, 01),\n#     ships=["Deep Space 9", "Defiant"],\n# )\n\nkirk.delete()\nCaptain.delete_key("key2")\n\nCaptain.get_all()\n# []\n\n# you can also save several at once for better speed\nCaptain.put_many([kirk, sisko])\n# [\n#     Captain(\n#         name="James T. Kirk", \n#         joined=datetime.date(2252, 01, 01), \n#         ships=["Enterprise", "Enterprise-A"],\n#         key="key1",\n#     ),\n#     Captain(\n#         name="Benjamin Sisko",\n#         joined=datetime.date(2350, 01, 01), \n#         ships=["Deep Space 9", "Defiant"],\n#         key="key2",\n#     ),\n# ]\n\n```\n\n## Save\n\nModels have the `.save()` method which will always behave as an upsert, updating a\nrecord if it has a key, otherwise creating it and setting a key. Deta has pure insert\nbehavior, but it\'s less performant. If you need it, please open a pull request.\n\n## Querying\n\nAll basic comparison operators are implemented to map to their equivalents as\n`(Model.field >= comparison_value)`. There is also a `.contains()` and `.not_contains()`\nmethod for strings and lists of strings, as well as a `.prefix()` method for strings.\nThere is also a `.range()` for number types that takes a lower and upper bound. You can\nalso use `&`  as AND and `|` as OR. ORs cannot be nested within ands, use a list of\noptions as comparison instead. You can use as many ORs as you want, as long as they\nexecute after the ANDs in the order of operations. This is due to how the Deta Base api\nworks.\n\n## Deta Base\n\nDirect access to the base is available in the dunder attribute `__db__`, though the\npoint is to avoid that.\n\n## Exceptions\n\n- `DetaError`: Base exception when anything goes wrong.\n- `ItemNotFound`: Fairly self-explanatory...\n- `NoProjectKey`: `PROJECT_KEY` env var has not been set correctly. See Deta\n  documentation.\n- `InvalidDetaQuery`: Something is wrong with queries. Make sure you aren\'t using\n  queries with unsupported types',
    'author': 'Rick Henry',
    'author_email': 'rickhenry@rickhenry.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/rickh94/odetam',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
